/*
 * Copyright (c) 1991-2023 by STEP Tools Inc. 
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation is hereby granted, provided that this copyright
 * notice and license appear on all copies of the software.
 * 
 * STEP TOOLS MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
 * SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT. STEP TOOLS
 * SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A
 * RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES.
 *
 * Author: David Loffredo (loffredo@steptools.com)
 */

#include <stp_schema.h>
#include "toolpath.h"

//------------------------------------------------------------
// Construct the various common SI and English units.
//
// These functions only create one instance of a given unit in a
// design.  The first time the unit is created, we save a shortcut in
// the nametable of the design.  This lets us quickly find the unit
// again if we need it, which keeps us from creating duplicate unit
// definitions.
//



//------------------------------------------------------------
// LENGTH UNITS
//

stp_named_unit * ToolpathMaker::make_length_unit()
{
    switch (unit_system) {
    case inch_deg_ipm_rpm:  	return make_inch_unit();
    case mm_deg_mmpm_rpm:   	return make_mm_unit();
    default:			return make_mm_unit();
    }
}

stp_named_unit * ToolpathMaker::make_mm_unit()
{
    const char * shortcut = "__CONSTANT UNIT millimetre";

    if (!f_design) return 0;

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = f_design-> findObject (shortcut);
    if (obj) return ROSE_CAST (stp_named_unit, obj);

    // Need a complex instance of both si and length unit
    stp_si_unit * u = pnewIn(f_design) stp_length_unit_and_si_unit;
    u-> name (stp_si_unit_name_metre);
    u-> prefix (stp_si_prefix_milli);

    f_design-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}


stp_named_unit * ToolpathMaker::make_inch_unit()
{
    const char * shortcut = "__CONSTANT UNIT inch";

    if (!f_design) return 0;

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = f_design-> findObject (shortcut);
    if (obj) return ROSE_CAST (stp_named_unit, obj);

    // Inch is defined as a conversion-based unit in terms of an
    // underlying SI unit.  Usually 25.4mm or 2.54cm.  Create the
    // measure with unit to indicate this.  
    //
    stp_measure_with_unit * mwu = pnewIn(f_design) 
	stp_length_measure_with_unit;
    mwu-> value_component (pnewIn(f_design) stp_measure_value);
    mwu-> value_component ()-> _length_measure (25.4);

    mwu-> unit_component (pnewIn(f_design) stp_unit);
    mwu-> unit_component ()-> _named_unit (make_mm_unit());


    // The unit itself needs a complex instance of both CBU and length
    // unit.  Conversion based unit also needs dimensional exponents
    // that show which basic quantities (length, time, mass, etc) the
    // unit describes.
    //
    stp_conversion_based_unit * u = pnewIn(f_design) 
	stp_conversion_based_unit_and_length_unit;

    u-> name ("inch");
    u-> conversion_factor (mwu);
    u-> dimensions (pnewIn(f_design) stp_dimensional_exponents);
    u-> dimensions()-> length_exponent(1);

    f_design-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}


//------------------------------------------------------------
// ANGLE UNITS
//

stp_named_unit * ToolpathMaker::make_angle_unit()
{
    return make_degree_unit();
}

stp_named_unit * ToolpathMaker::make_radian_unit()
{
    const char * shortcut = "__CONSTANT UNIT radian";

    if (!f_design) return 0;

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = f_design-> findObject (shortcut);
    if (obj) return ROSE_CAST (stp_named_unit, obj);

    // Need a complex instance of both si and plane angle unit
    stp_si_unit * u = pnewIn(f_design) stp_plane_angle_unit_and_si_unit;
    u-> name (stp_si_unit_name_radian);
    u-> prefix (stp_si_prefix_NULL);

    f_design-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}



stp_named_unit * ToolpathMaker::make_degree_unit()
{
    const char * shortcut = "__CONSTANT UNIT degree";

    if (!f_design) return 0;

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = f_design-> findObject (shortcut);
    if (obj) return ROSE_CAST (stp_named_unit, obj);

    // Degree is defined as a conversion-based unit in terms of the
    // underlying radian SI unit.  1 Degree == PI/180 radians.  Create
    // the measure with unit to indicate this.
    //
    stp_measure_with_unit * mwu = pnewIn(f_design) 
	stp_plane_angle_measure_with_unit;
    mwu-> value_component (pnewIn(f_design) stp_measure_value);
    mwu-> value_component ()-> _plane_angle_measure (0.01745329252);

    mwu-> unit_component (pnewIn(f_design) stp_unit);
    mwu-> unit_component ()-> _named_unit (make_radian_unit());


    // The unit itself needs a complex instance of both CBU and plane
    // angle unit.  Conversion based unit also needs dimensional
    // exponents but these are all zero for angles.
    //
    stp_conversion_based_unit * u = pnewIn(f_design) 
	stp_conversion_based_unit_and_plane_angle_unit;

    u-> name ("degree");
    u-> conversion_factor (mwu);
    u-> dimensions (pnewIn(f_design) stp_dimensional_exponents);

    f_design-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}



stp_named_unit * ToolpathMaker::make_steradian_unit()
{
    const char * shortcut = "__CONSTANT UNIT steradian";

    if (!f_design) return 0;

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = f_design-> findObject (shortcut);
    if (obj) return ROSE_CAST (stp_named_unit, obj);

    // Need a complex instance of both si and length unit
    stp_si_unit * u = pnewIn(f_design) stp_si_unit_and_solid_angle_unit;
    u-> name (stp_si_unit_name_steradian);
    u-> prefix (stp_si_prefix_NULL);

    f_design-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}




//------------------------------------------------------------
// TIME UNITS
//

stp_named_unit * ToolpathMaker::make_second_unit() 
{
    const char * shortcut = "__CONSTANT UNIT second";

    if (!f_design) return 0;

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = f_design-> findObject (shortcut);
    if (obj) return ROSE_CAST (stp_named_unit, obj);

    // Need a complex instance of both si and length unit
    stp_si_unit * u = pnewIn(f_design) stp_si_unit_and_time_unit;
    u-> name (stp_si_unit_name_second);
    u-> prefix (stp_si_prefix_NULL);

    f_design-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}

stp_named_unit * ToolpathMaker::make_minute_unit()
{
    const char * shortcut = "__CONSTANT UNIT minute";

    if (!f_design) return 0;

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = f_design-> findObject (shortcut);
    if (obj) return ROSE_CAST (stp_named_unit, obj);

    // minute is a conversion-based unit based in seconds
    stp_measure_with_unit * mwu = pnewIn(f_design) stp_time_measure_with_unit;
    mwu-> value_component (pnewIn(f_design) stp_measure_value);
    mwu-> value_component ()-> _time_measure (60);

    mwu-> unit_component (pnewIn(f_design) stp_unit);
    mwu-> unit_component ()-> _named_unit (make_second_unit());


    // The unit itself needs a complex instance of both CBU and length
    // unit.  Conversion based unit also needs dimensional exponents
    // that show which basic quantities (length, time, mass, etc) the
    // unit describes.
    //
    stp_conversion_based_unit * u = pnewIn(f_design)
	stp_conversion_based_unit_and_time_unit;

    u-> name ("minute");
    u-> conversion_factor (mwu);
    u-> dimensions (pnewIn(f_design) stp_dimensional_exponents);
    u-> dimensions()-> time_exponent(1);

    f_design-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}



//------------------------------------------------------------
// LINEAR SPEED UNITS
//
static stp_derived_unit * make_derived_unit_base(
    RoseDesign * d,
    const char * name
    )
{
    stp_derived_unit * u = pnewIn(d) stp_derived_unit;
    stp_name_attribute * nm = pnewIn(d) stp_name_attribute;

    nm-> attribute_value (name);
    nm-> named_item (pnewIn(d) stp_name_attribute_select);
    nm-> named_item()-> _derived_unit(u);
    u-> elements (pnewIn(d) SetOfstp_derived_unit_element);
    return u;
}

static void add_derived_unit_term(
    stp_derived_unit * u,
    stp_named_unit * term,
    double exp
    )
{
    if (!u || !term) return;
    stp_derived_unit_element * elem = 
	pnewIn(u-> design()) stp_derived_unit_element;

    elem-> unit (term);
    elem-> exponent (exp);

    u-> elements()-> add(elem);
}


stp_derived_unit * ToolpathMaker::make_feed_unit()
{
    switch (unit_system) {
    case inch_deg_ipm_rpm:  	return make_ipm_unit();
    case mm_deg_mmpm_rpm:   	return make_mmpm_unit();
    default:			return make_mmpm_unit();
    }
}

stp_derived_unit * ToolpathMaker::make_mmpm_unit()
{
    const char * shortcut = "__CONSTANT UNIT mm-per-minute";

    if (!f_design) return 0;

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = f_design-> findObject (shortcut);
    if (obj) return ROSE_CAST (stp_derived_unit, obj);

    stp_derived_unit * u = make_derived_unit_base(f_design, "millimetre/minute");
    add_derived_unit_term (u, make_mm_unit(), 1);
    add_derived_unit_term (u, make_minute_unit(), -1);

    f_design-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}



stp_derived_unit * ToolpathMaker::make_ipm_unit()
{
    const char * shortcut = "__CONSTANT UNIT inch-per-minute";

    if (!f_design) return 0;

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = f_design-> findObject (shortcut);
    if (obj) return ROSE_CAST (stp_derived_unit, obj);

    stp_derived_unit * u = make_derived_unit_base(f_design, "inch/minute");
    add_derived_unit_term (u, make_inch_unit(), 1);
    add_derived_unit_term (u, make_minute_unit(), -1);

    f_design-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}




//------------------------------------------------------------
// ROTATIONAL SPEED UNITS
//


stp_derived_unit * ToolpathMaker::make_spindle_unit()
{
    return make_rpm_unit();
}


stp_named_unit * ToolpathMaker::make_revolution_unit()
{
    const char * shortcut = "__CONSTANT UNIT revolution";

    if (!f_design) return 0;

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = f_design-> findObject (shortcut);
    if (obj) return ROSE_CAST (stp_named_unit, obj);

    stp_context_dependent_unit * u = pnewIn(f_design) 	
	stp_context_dependent_unit;

    u-> name ("revolution");
    u-> dimensions (pnewIn(f_design) stp_dimensional_exponents);

    f_design-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}


stp_derived_unit * ToolpathMaker::make_rps_unit()
{
    const char * shortcut = "__CONSTANT UNIT rev-per-second";

    if (!f_design) return 0;

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = f_design-> findObject (shortcut);
    if (obj) return ROSE_CAST (stp_derived_unit, obj);

    stp_derived_unit * u = make_derived_unit_base(f_design, "revolution/second");
    add_derived_unit_term (u, make_revolution_unit(), 1);
    add_derived_unit_term (u, make_second_unit(), -1);

    f_design-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}


stp_derived_unit * ToolpathMaker::make_rpm_unit()
{
    const char * shortcut = "__CONSTANT UNIT rev-per-minute";

    if (!f_design) return 0;

    /* Use shortcut to unit if we have already created one */
    RoseObject * obj = f_design-> findObject (shortcut);
    if (obj) return ROSE_CAST (stp_derived_unit, obj);

    stp_derived_unit * u = make_derived_unit_base(f_design, "revolution/minute");
    add_derived_unit_term (u, make_revolution_unit(), 1);
    add_derived_unit_term (u, make_minute_unit(), -1);

    f_design-> addName (shortcut, u);  // save shortcut for future calls
    return u;
}
